﻿//  Memory_Helper.cpp
//
//  Galaxy Group
//  Minsk, Slavic Rus
//  https://www.councils.world/
//  Kozlov Sergey Aleksandrovitsch
//  Since 2025 year.

#include "pch.h"

#include <string>
#include <process.h>

#include "Memory_Helper.h"

#include "Memory_Helper_Worker_Exception.h"

#ifdef USING_MFC

#ifdef _DEBUG
#define new DEBUG_NEW
#endif

#endif

using namespace std;

Galaxy_Group_Star::Memory_Helper::Memory_Helper_Worker *mhw = nullptr;

int main()
{
    return entrance();
}

extern "C" MEMORY_HELPER_API int entrance()
{
    int nRetCode = 0;

    HMODULE hModule = ::GetModuleHandle(nullptr);

    if (hModule != nullptr)
    {
#ifdef USING_MFC

        if (!AfxWinInit(hModule, nullptr, ::GetCommandLine(), 0))
        {
            wprintf(L"Fatal Error: MFC initialization failed.\n");

            nRetCode = 1;
        }
        else
#endif
        {
            if (mhw == nullptr)
            {
                try
                {
                    mhw = new Galaxy_Group_Star::Memory_Helper::Memory_Helper_Worker();

                    if (mhw != nullptr)
                    {
                        mhw->Work();
                    }
                }
                catch (Galaxy_Group_Star::Exception::Memory_Helper_Worker_Exception* mhwe)
                {
                    if (mhwe != nullptr)
                    {
                        wprintf(mhwe->What().c_str());
                    }
                    else
                    {
                        wprintf(L"Fatal error in Memory_Helper_Worker.\n");
                    }
                    nRetCode = 1;
                }
                catch (...)
                {
                    wprintf(L"Unknown error in Memory_Helper_Worker.\n");

                    nRetCode = 1;
                }
            }
        }
    }
    else
    {
        wprintf(L"Fatal Error: GetModuleHandle failed.\n");

        nRetCode = 1;
    }

    return nRetCode;
}

extern "C" MEMORY_HELPER_API bool set_call_back_alarm_function(call_back_alarm_function_type* p)
{
    if (mhw != nullptr)
    {
        mhw->Set_Call_Back_Alarm_Function(p);

        return true;
    }
    else
    {
        return false;
    }
}

extern "C" MEMORY_HELPER_API bool terminate_memory_helper()
{
    if (mhw != nullptr)
    {
        mhw->Terminate();

        for (; true;)
        {
            if (mhw == nullptr)
            {
                return false;
            }

            if (mhw->Get_Termanation_Allowed())
            {
                break;
            }

            Sleep(1);
        }

        if (mhw != nullptr)
        {
            delete mhw;

            mhw = nullptr;
        }

        return true;
    }
    else
    {
        return false;
    }
}

namespace Galaxy_Group_Star
{
    namespace Memory_Helper
    {
        Memory_Helper_Worker::Memory_Helper_Worker()
            : Call_Back_Alarm_Function(nullptr), do_termanate(false), termanation_allowed(false), working_thread(0)
        {

        }

        Memory_Helper_Worker::Memory_Helper_Worker(const Memory_Helper_Worker& p)
        {
            Call_Back_Alarm_Function = p.Call_Back_Alarm_Function;
            
            do_termanate = p.do_termanate;

            termanation_allowed = p.termanation_allowed;

            working_thread = p.working_thread;
        }

        Memory_Helper_Worker& Memory_Helper_Worker::operator=(const Memory_Helper_Worker& p)
        {
            Call_Back_Alarm_Function = p.Call_Back_Alarm_Function;

            do_termanate = p.do_termanate;

            termanation_allowed = p.termanation_allowed;

            working_thread = p.working_thread;

            return *this;
        }

        Memory_Helper_Worker::~Memory_Helper_Worker()
        {
        }

        void Memory_Helper_Worker::Work()
        {
            Memory_Helper_Worker_Thread_Function_Parameters* p = new Memory_Helper_Worker_Thread_Function_Parameters();

            if (p != nullptr)
            {
                p->memory_helper_worker = this;

                working_thread = _beginthread(&Memory_Helper_Worker_Thread_Function, 0, p);
            }            
        }

        void Memory_Helper_Worker::Check()
        {
            if (Call_Back_Alarm_Function != nullptr)
            {
                if (Is_Memory_Alarm_Occured())
                {
                    Call_Back_Alarm_Function();
                }
            }
        }

        void Memory_Helper_Worker::Set_Call_Back_Alarm_Function(call_back_alarm_function_type* p)
        {
            Call_Back_Alarm_Function = p;
        }

        void Memory_Helper_Worker::Terminate()
        {
            do_termanate = true;
        }

        bool Memory_Helper_Worker::Is_Memory_Alarm_Occured()
        {
            return Control_Memory();
        }

        bool Memory_Helper_Worker::Control_Memory()
        {
            bool memory_control_alarm = false;

            MEMORYSTATUSEX memory_status;

            memset(&memory_status, 0, sizeof(MEMORYSTATUSEX));
            memory_status.dwLength = sizeof(MEMORYSTATUSEX);

            if (GlobalMemoryStatusEx(&memory_status))
            {
                DWORDLONG total_memory = memory_status.ullTotalPhys;
                DWORDLONG available_memory = memory_status.ullAvailPhys;

                if (total_memory != 0)
                {
                    const DWORDLONG percents_of_memory_allowed = 10;

                    if (available_memory <= total_memory / 100 * percents_of_memory_allowed)
                    {
                        memory_control_alarm = true;

                        //throw Galaxy_Group_Star::Exception::Memory_Helper_Worker_Exception(std::wstring(L"Available less than 10 % of total memory - processing (decoding) will be stopped.\n"));
                    }
                    else
                    {
                        memory_control_alarm = false;
                    }
                }
                else
                {
                    throw Galaxy_Group_Star::Exception::Memory_Helper_Worker_Exception(std::wstring(L"Failed to retrieve memory status: installed zero bytes of memory.\n"));
                }
            }
            else
            {
                throw Galaxy_Group_Star::Exception::Memory_Helper_Worker_Exception(std::wstring(L"Failed to retrieve memory status.\n"));
            }

            return memory_control_alarm;
        }

        bool Memory_Helper_Worker::Get_Termanation_Allowed()
        {
            return termanation_allowed;
        }

        void Memory_Helper_Worker::Do_Work()
        {
            for (; !do_termanate;)
            {
                Sleep(Memory_Helper_Worker_Timer_Event_Delay);

                Check();
            }

            termanation_allowed = true;
        }

        void _cdecl Memory_Helper_Worker_Thread_Function(void* lp)
        {
            if (lp != nullptr)
            {
                Memory_Helper_Worker_Thread_Function_Parameters* mhwtp = (Memory_Helper_Worker_Thread_Function_Parameters*)(lp);

                if (mhwtp != nullptr)
                {
                    Memory_Helper_Worker* local_mhw = mhwtp->memory_helper_worker;

                    if (local_mhw != nullptr)
                    {
                        local_mhw->Do_Work();
                    }
                }

                delete lp;
            }
        }
    }
}